from ortools.linear_solver import pywraplp
from math import ceil, floor
from pool import Pool

def monotier(startersPerTeam, poolSize, nbTeams, tier):
    solver = pywraplp.Solver.CreateSolver("SCIP")

    if not solver:
        return
    nbPools = startersPerTeam * nbTeams / poolSize
    if (int(nbPools) != nbPools):
        print("Number of pools is not an integer.")
        return 
    
    nbPools = int(nbPools)
    
    #Variables
    #solo
    solo = {}
    for p in range(nbPools):
        for i in range(nbTeams):
            solo[(p, i)] = solver.IntVar(0, 1, "")

    #duo
    duo = {}
    for p in range(nbPools):
        for i in range(nbTeams - 1):        
            for j in range(i + 1, nbTeams):
                duo[(p, i, j)] = solver.IntVar(0, 1, "")

    #Constraints
    #Starters per Team
    for t in range(nbTeams):
        solver.Add(sum(solo[p, t] for p in range(nbPools)) == startersPerTeam)
    
    #Pool size
    for p in range(nbPools):
        solver.Add(sum(solo[p, t] for t in range(nbTeams)) == poolSize)
    
    #Link duo and solo
    for p in range(nbPools):
        for t1 in range(nbTeams - 1):
            for t2 in range(t1 + 1, nbTeams):
                solver.Add(solo[p, t1] + solo[p, t2] - 2 * duo[p, t1, t2] <= 1)
                solver.Add(2 * duo[p, t1, t2] <= solo[p, t1] + solo[p, t2])

    m = startersPerTeam * (poolSize - 1) / (nbTeams - 1)

    #Every team faces each other at least a certain number of times
    for t1 in range(nbTeams - 1):
        for t2 in range(t1 + 1, nbTeams):
            solver.Add(sum(duo[(p, t1, t2)] for p in range(nbPools)) >= floor(m))

    #Every team faces each other at most a certain number of times
    for t1 in range(nbTeams - 1):
        for t2 in range(t1 + 1, nbTeams):
            solver.Add(sum(duo[(p, t1, t2)] for p in range(nbPools)) <= ceil(m))

    #Breaking symetries
    #Disjoint first pools
    q = 0
    for t in range(floor(startersPerTeam / 2)):
        for _ in range (startersPerTeam - 2 * t):
            solver.Add(solo[q, t] == 1)
            q += 1
    
    for p in range(q, nbPools):
        solver.Add(solo[p, 0] == 0)
        solver.Add(duo[p, 0, nbTeams-1] == 0)
        for t1 in range(1, nbTeams-1):
            solver.Add(duo[p, 0, t1] == 0)

    a = startersPerTeam
    for t in range(1, floor(startersPerTeam / 2)):
        a += startersPerTeam - 2
        for p in range(a, q):
            solver.Add(solo[p, t] == 0)
            solver.Add(duo[p, t, nbTeams-1] == 0)
            for t1 in range(t+1, nbTeams - 1):
                solver.Add(duo[p, t, t1] == 0)

    #Force first pool
    for t1 in range(poolSize - 2):
        for t2 in range(t1 + 1, poolSize - 1):
            for t3 in range(t2 + 2, poolSize):
                solver.Add(solo[0, t3] == 1)
                solver.Add(duo[0, t2, t3] == 1)
            for t3 in range(poolSize, nbTeams):
                solver.Add(solo[0, t3] == 0)
                solver.Add(duo[0, t2, t3] == 0)
        for t2 in range(poolSize - 1, nbTeams - 1):
            for t3 in range(t2 + 1, nbTeams):
                solver.Add(duo[0, t2, t3] == 0)

    #Force second pools
    a = min(q + startersPerTeam - 1, nbTeams)
    for p in range(1, startersPerTeam):
        for t in range(q, a):
            solver.Add(solo[p, t] == 1)
            solver.Add(duo[p, 0, t] == 1)

    #Ban last pools
    a = 0
    for t in range(1, startersPerTeam):
        a = a + 2 * t
        for p in range(q, nbPools - a):
            ##solver.Add(solo[p, t] == 0)
            if (t < nbTeams - 1):
                for t1 in range(t + 1, nbTeams):
                    solver.Add(duo[p, t, t1] == 0)

    print(f"Solving with {solver.SolverVersion()}")
    status = solver.Solve()
    if status != solver.OPTIMAL:
        print("The problem does not have an optimal solution!")
        if status == solver.FEASIBLE:
            print("A potentially suboptimal solution was found.")
        else:
            print("The solver could not solve the problem.")
            exit(1)

    pools = [] 
    for p in range(nbPools):
        pool = Pool(p + 1, tier)
        for t in range(nbTeams):
            if solo[p, t].solution_value() == 1:
                pool.addPlayer(t)
        pools.append(pool)
    return pools

def multiTiers(startersPerFormat, poolSize, nbTeams, tiers):
    solver = pywraplp.Solver.CreateSolver("SCIP")

    if not solver:
        return

    if (int(nbTeams / poolSize) != nbTeams / poolSize):
        print("Number of pools is not an integer.")
        return 
    
    nbPools = int(sum(startersPerFormat) * nbTeams / poolSize)
    nbFormat = len(startersPerFormat)

    startingIndexPerFormat = [0]
    index = 0
    for starters in startersPerFormat:
        index += int(starters * nbTeams / poolSize)
        startingIndexPerFormat.append(index)
    
    #Variables
    #solo
    solo = {}
    for p in range(nbPools):
        for i in range(nbTeams):
            solo[(p, i)] = solver.IntVar(0, 1, "")

    #duo
    duo = {}
    for p in range(nbPools):
        for i in range(nbTeams - 1):        
            for j in range(i + 1, nbTeams):
                duo[(p, i, j)] = solver.IntVar(0, 1, "")
    
    #Constraints
    #Starters per Team
    for f in range(nbFormat):
        for t in range(nbTeams):
            solver.Add(sum(solo[p, t] for p in range(startingIndexPerFormat[f], startingIndexPerFormat[f+1])) == startersPerFormat[f])
    
    #Pool size
    for p in range(nbPools):
        solver.Add(sum(solo[p, t] for t in range(nbTeams)) == poolSize)
    
    #Link duo and solo
    for p in range(nbPools):
        for t1 in range(nbTeams - 1):
            for t2 in range(t1 + 1, nbTeams):
                solver.Add(solo[p, t1] + solo[p, t2] - 2 * duo[p, t1, t2] <= 1)
                solver.Add(2 * duo[p, t1, t2] <= solo[p, t1] + solo[p, t2])

    m = sum(startersPerFormat) * (poolSize - 1) / (nbTeams - 1)

    #Every team faces each other at least a certain number of times
    for t1 in range(nbTeams - 1):
        for t2 in range(t1 + 1, nbTeams):
            solver.Add(sum(duo[(p, t1, t2)] for p in range(nbPools)) >= floor(m))

    #Every team faces each other at most a certain number of times
    for t1 in range(nbTeams - 1):
        for t2 in range(t1 + 1, nbTeams):
            solver.Add(sum(duo[(p, t1, t2)] for p in range(nbPools)) <= ceil(m))
                
    #Breaking symetries
    #Disjoint first pools
    t = 1
    for f in range(nbFormat):
        if startersPerFormat[f] == 1:
            p = startingIndexPerFormat[f]
            for t in range(nbTeams):
                for _ in range(poolSize):
                    solver.Add(solo[(p, t)] == 1)
                p += 1
            break
        
    print(f"Solving with {solver.SolverVersion()}")
    status = solver.Solve()
    if status != solver.OPTIMAL:
        print("The problem does not have an optimal solution!")
        if status == solver.FEASIBLE:
            print("A potentially suboptimal solution was found.")
        else:
            print("The solver could not solve the problem.")
            exit(1)

    pools = [] 
    for f in range(nbFormat):
        for p in range(startingIndexPerFormat[f + 1] - startingIndexPerFormat[f]):
            pool = Pool(p + 1, tiers[f])
            for t in range(nbTeams):
                if solo[startingIndexPerFormat[f] + p, t].solution_value() == 1:
                    pool.addPlayer(t)
            pools.append(pool)
    return pools

def printPools(pools):
    for pool in pools:
        print(pool)

if __name__ == "__main__":
    #printPools(monotier(6, 4, 12, "SV ZU"))
    printPools(multiTiers([5, 1, 1, 1, 1, 1], 4, 20, ["SV OU", "SS OU", "SM OU", "ORAS OU", "SV Ubers", "SV UU"]))